/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: simulation_common.c,v $
*	$Revision: 1.65 $
*	$Date: 2008/06/02 16:27:51 $
*	$Author: i.zapreev $
*
*	MRMC is a model checker for discrete-time and continuous-time Markov reward models.
*	It supports reward extensions of PCTL and CSL (PRCTL and CSRL), and allows for the
*	automated verification of properties concerning long-run and instantaneous rewards
*	as well as cumulative rewards.
*
*	Copyright (C) The University of Twente, 2004-2007.
*	Authors: Ivan Zapreev
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
*	Our contacts:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: This is a header file for the simulation engine
*	here we intend to define the common functions and data structures.
*/

#include "simulation_common.h"

#include "parser_to_tree.h"
#include "simulation.h"
#include "simulation_utils.h"
#include "sample_vec.h"
#include "gsl/gsl_cdf.h"

/****************************************************************************/
/********THE COMMON FUNCTION FOR INVOCATION THE SIMULATION PROCEDURE ********/
/****************************************************************************/

/**
* This function is responsible for invoking the model procedure either
* for one or for all initial states the invoked finction for model checking
* one initial state is passed as an argument pFMCOneState
* @param pStateSpace the state-space matrix
* @param confidence this is is confidence we should use for simulations.
*	Note that since the formulas can be nested one can not use the
*	overall confidence set in simulation.h
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param initial_state the initial state index if isSimOneInitState == TRUE
*
* NOTE: Below we have pointers to the pointers, these are going to be the return values
*
* @param ppYesBitSet this bitset is going to be filled with the states which
*			satisfy the formula
* @param ppNoBitSet this bitset is going to be filled with the states which
*			do not satisfy the formula
* @param ppCiLeftBorders will contain the left conf. int. borders
* @param ppCiRightBorders will contain the right conf. int. borders
* @param pResultSize the pointer to the size of arrays *ppCiLeftBorders and *ppCiRightBorders
* @param pValidInitialStates the bitset containing all allowed initial
* @param pMaxNumUsedObserv the return-value variable that will store the maximum number
* 			(over all initial states) of used observations
* @param pFMCOneState the pointer to a function which model checks one initial state
* WARNING: The parameter pFMCOneState should be the last one before "..."
* @param ... the extra arguments needed for a specific model-checking procedure
*/
void modelCheckStatesCommon( const sparse * pStateSpace, const double confidence, bitset ** ppYesBitSet,
				bitset ** ppNoBitSet, double ** ppCiLeftBorders, double ** ppCiRightBorders,
				int * pResultSize, const int comparator, const double prob_bound,
				const int initial_state, const BOOL isSimOneInitState,
				const bitset * pValidInitialStates, unsigned int *pMaxNumUsedObserv,
				TPFunctMCOneState pFMCOneState, ... ){
	/* Will contain the extra arguments of the function pFMCOneState */
	va_list fMCOneStateArguments;
	/* Obtain the width of the indifference region */
	const double indiff_width = getSupIndifferenceWidth();
	/* Will store the number of used observations */
	unsigned int numUsedObserv = 0;

	/* Allocate the resulting bitsets */
	const int state_space_size = pStateSpace->rows;
	*ppYesBitSet = get_new_bitset( state_space_size );
	*ppNoBitSet = get_new_bitset( state_space_size );
	
	/* Reset the maximum number of used observations. */
	*pMaxNumUsedObserv = 0;
	
	if( ( *ppYesBitSet != NULL ) && ( *ppYesBitSet != NULL ) ){
		/* Depending on whether there is one initial state only or not. */
		if( isSimOneInitState ) {
			/* Check for the valid initial state */
			IF_SAFETY( ( initial_state >= 0) || ( initial_state < state_space_size ) )
				/* If it is a valid state */
				if( get_bit_val( pValidInitialStates, initial_state ) ){
					/* Allocate the return data */
					*pResultSize = 1;
					*ppCiLeftBorders = (double * ) calloc( 1, sizeof(double) );
					*ppCiRightBorders = (double * ) calloc( 1, sizeof(double) );
					
					/* Obtain the extra arguments of the function pFMCOneState */
					va_start( fMCOneStateArguments, pFMCOneState );

					/* Reset the observations counter for one initial state */
					numUsedObserv = 0;
					
					/* Model check the state */
					pFMCOneState( pStateSpace, initial_state, indiff_width, confidence,
							comparator, prob_bound, 0, *ppCiLeftBorders, *ppCiRightBorders,
							*ppYesBitSet, *ppNoBitSet, &numUsedObserv, fMCOneStateArguments );
					
					/* Since there is just one initial state: maxNumUsedObserv = numUsedObserv */
					*pMaxNumUsedObserv = numUsedObserv;
					
					/* Complete the argument list usage */
					va_end( fMCOneStateArguments );
				}else{
					*pResultSize = 0;
				}
			ELSE_SAFETY
				printf("ERROR: An ivalid initial state %d, should be >= %d and <= %d.\n",
					initial_state+1, 1, state_space_size );
				exit(1);
			ENDIF_SAFETY
		} else {
			int curr_state = -1;
			
			/* Allocate the return data */
			*pResultSize = state_space_size;
			*ppCiLeftBorders = (double * ) calloc( state_space_size, sizeof(double) );
			*ppCiRightBorders = (double * ) calloc( state_space_size, sizeof(double) );
			
			/* Work out all the transient states */
			while( ( curr_state = get_idx_next_non_zero( pValidInitialStates, curr_state  )) != -1 ){
				/* Obtain the extra arguments of the function pFMCOneState */
				va_start( fMCOneStateArguments, pFMCOneState );
				
				/* Reset the observations counter for one initial state */
				numUsedObserv = 0;
				
				/* Model check the state */
				pFMCOneState( pStateSpace, curr_state, indiff_width, confidence, comparator,
						prob_bound, curr_state, *ppCiLeftBorders, *ppCiRightBorders,
						*ppYesBitSet, *ppNoBitSet, &numUsedObserv, fMCOneStateArguments );
				
				/* Take the muximum between the number of observations used now and before*/
				if( numUsedObserv > *pMaxNumUsedObserv ){
					*pMaxNumUsedObserv = numUsedObserv;
				}
				
				/* printf("The number of observations used to simulate state %d is %u.\n", curr_state+1, numUsedObserv ); */
				
				/* Complete the argument list usage */
				va_end( fMCOneStateArguments );
			}
		}
	} else {
		printf("ERROR: Failed to allocate data structures for the simulation engine.\n");
		exit(1);
	}
}

/****************************************************************************/
/********THE COMMON FUNCTION FOR SIMULATING UNBOUNDED UNTIL CTMC/DTMC********/
/****************************************************************************/

/**
* Model checks the unbounded-until operator for one initial state only.
* Note that this is a univesral procedure that assumes that we work with the embedded DTMC.
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC with good and bad
*			states made absorbing
* @param initial_state the initial state for simulations
* @param indiff_width the width of the indifference region
* @param confidence the desired confidence of the answer
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param arr_index indicates the index in arrays pCiLeftBorders and pCiRightBorders
*			that corresponds to the initial_state
*
* NOTE: The parameters below will contain the return values
*
* @param pCiLeftBorders for storing the left conf. int. border
* @param pCiRightBorders for storing the right conf. int. border
* @param pYesBitSet contains curr_state if it satisfies the formula
* @param pNoBitSet contains curr_state if it does not satisfy the formula
* @param pNumUsedObserv the pointer to the return-value variable that will store the
*		number of observations used for simulating this one state.
* @param pTransAndGoodStatesArguments contain the two bitsets in the following order:
*		1. pTransientStates the bitsets containing all the transient states
*		2. pGoodStates the bitsets containing all the good absorbing states
*/
void modelCheckOneStateUUCommon( const sparse* pStateSpace, const int initial_state, const double indiff_width,
					const double confidence, const int comparator, const double prob_bound,
					const int arr_index, double *pCiLeftBorders, double *pCiRightBorders,
					bitset * pYesBitSet, bitset * pNoBitSet, unsigned int *pNumUsedObserv,
					va_list pTransAndGoodStatesArguments ){
	/* Contain the Transient states and also the Good absorbing states respectively */
	bitset * pTransientStates = va_arg( pTransAndGoodStatesArguments, bitset * );
	bitset * pGoodStates = va_arg( pTransAndGoodStatesArguments, bitset * );
	
	TV_LOGIC mc_result = TVL_NN;
	double leftBorderOne = 0.0, rightBorderOne = 1.0, leftBorderTwo = 0.0, rightBorderTwo = 1.0;
	
	/* Compute the general confidence zeta value: */
	const double gen_conf_zeta = gsl_cdf_ugaussian_Pinv( sqrt( confidence ) );
	/* NOTE: The confidence we have is used to derive z for the left and the */
	/* right border BUT the TRUE/FALSE answer is based only on one border of */
	/* the conf. int., i.e. we could use gsl_cdf_ugaussian_Pinv( confidence ) */
	/* for model checking, but then again we have a different confidence for the */
	/* produced conf. int. this is bad because then the user does not know that. */
	
	/* Initialize the sample size parameters */
	int sample_size = getSimMinSampleSize();
	const int sample_size_step = getSimSampleSizeStep();
	const int max_sample_size = getSimMaxSampleSize();
	
	/* Initialize the simulation depth parameters */
	int simulation_depth = getSimMinSimulationDepth();
	const int simulation_depth_step = getSimSimulationDepthStep();
	const int max_simulation_depth = getSimMaxSimulationDepth();
	
	/* The odd/even iteration and invalid conf.int. indicators */
	BOOL isOdd = TRUE, isInvalidCI = FALSE;
	
	/* Allocate and simulate the initial sample vectors */
	PTSampleVecUntil pSampleVecUntilOne = allocateSampleVectorUntil( sample_size, initial_state );
	PTSampleVecUntil pSampleVecUntilTwo = allocateSampleVectorUntil( sample_size, initial_state );
	
	/* Simulate observations in depth */
	simulateSampleVectorUnbUntilDTMC( pStateSpace, pSampleVecUntilOne, sample_size, simulation_depth,
						pGoodStates, pTransientStates );
	simulateSampleVectorUnbUntilDTMC( pStateSpace, pSampleVecUntilTwo, sample_size, simulation_depth,
						pGoodStates, pTransientStates );
	
	/* Perform the main simulation cycle */
	do{
		/* Choose what to increment  */
		if( ( isOdd || ( sample_size >= max_sample_size ) ) &&
			( simulation_depth < max_simulation_depth ) && !isInvalidCI ){
			/*One could also check for the following: */
			/*( pSampleVecUntilOne->sum_trans + pSampleVecUntilTwo->sum_trans ) > 0 */
			/*but it seems to cause infinite looping some times and*/
			/*there is not real need for checking this condition, because */
			/*in this case we mostly skip the iteration any ways.*/
			increment( &simulation_depth, simulation_depth_step, max_simulation_depth );
		} else {
			increment( &sample_size, sample_size_step, max_sample_size );
		}
		
		/* Simulate the samples */
		simulateSampleVectorUnbUntilDTMC( pStateSpace, pSampleVecUntilOne, sample_size, simulation_depth,
							pGoodStates, pTransientStates );
		simulateSampleVectorUnbUntilDTMC( pStateSpace, pSampleVecUntilTwo, sample_size, simulation_depth,
							pGoodStates, pTransientStates );
		
		/* Compute the Conf Int borders */
		computeBordersUU( gen_conf_zeta, &leftBorderOne, &rightBorderOne, pSampleVecUntilOne, AGRESTI_COULL_CONF_INT );
		computeBordersUU( gen_conf_zeta, &leftBorderTwo, &rightBorderTwo, pSampleVecUntilTwo, AGRESTI_COULL_CONF_INT );
		
		/* Test the produced conf. int. for having a non-zero intersection */
		/* Note that leftBorderOne <= rightBorderOne and leftBorderTwo <= rightBorderTwo */
		isInvalidCI = ( rightBorderOne < leftBorderTwo ) || ( rightBorderTwo < leftBorderOne);
		
		/* Check the probability constraint */
		/* WARNING: We can not do any sort of sample swapping here! */
		if( !isInvalidCI ){
			mc_result = checkBoundVSConfInt( comparator, prob_bound, leftBorderOne, rightBorderTwo, indiff_width );
		}
		isOdd = ! isOdd;
	}while( ( mc_result == TVL_NN ) &&
			( ( ( simulation_depth < max_simulation_depth ) && ! isInvalidCI ) ||
				( sample_size < max_sample_size ) ) );
	
	/* Update the pCiLeftBorders and pCiRightBorders */
	pCiLeftBorders[arr_index] = leftBorderOne;
	pCiRightBorders[arr_index] = rightBorderTwo;
	
	/* printf("Result: %i, borders are: [%lf, %lf], prob bound %d %lf.\n", mc_result, */
	/* 	pCiLeftBorders[arr_index], pCiRightBorders[arr_index], comparator, prob_bound); */
	
	/* Update the pYesBitSet or pNoBitSet if we have a definite answer */
	markYesNoSetEntree(mc_result, initial_state, pYesBitSet, pNoBitSet);
	
	/* printf("mc_result = %d, sample_size = %d and simulation_depth = %d\n", */
	/* 	mc_result, sample_size, simulation_depth); */
	/* printSampleVectorUntil(pSampleVecUntilOne); */
	/* printSampleVectorUntil(pSampleVecUntilTwo); */

	/* Report the number of states visited while this simulation run */
	*pNumUsedObserv = ( (PTSampleVec) pSampleVecUntilOne )->num_visited_states +
				( (PTSampleVec) pSampleVecUntilTwo )->num_visited_states;
	
	/* Free the sample vectors */
	freeSampleVectorUntil( pSampleVecUntilOne );
	freeSampleVectorUntil( pSampleVecUntilTwo );
}
